/*

Mures
Copyright (C) 2001 Adam D'Angelo

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact information:

Adam D'Angelo
dangelo@ntplx.net
P.O. Box 1155
Redding, CT 06875-1155
USA

*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "SDL.h"

#include "protocol.h"
#include "types.h"
#include "sim.h"
#include "map.h"
#include "input.h"
#include "keyboard.h"
#include "output.h"
#include "ai.h"
#include "client.h"
#include "server.h"
#include "main.h"
#include "lib.h"

struct
{
  int server;
  int output;
  int local_player_count;
  int local_ai_count;
  int client;
  int fullscreen;
  int teams;
  char *host;
} global_option;

void add_local_players()
{
  int i,j;

  for(i=0; i<global_option.local_player_count; i++)
    for(j=0; j<MAX_PLAYER; j++)
      if(have_local_player[j])
	if(local_player_type[j] == UNKNOWN_PLAYER) {
	  local_player_type[j] = KEYBOARD_PLAYER;
	  break;
	}

  for(i=0; i<global_option.local_ai_count; i++)
    for(j=0; j<MAX_PLAYER; j++)
      if(have_local_player[j])
	if(local_player_type[j] == UNKNOWN_PLAYER) {
	  local_player_type[j] = AI_PLAYER;
	  break;
	}
}

Uint32 wait_frame(void)
{
  static Uint32 next_tick = 0;
  static Uint32 last_tick = 0;
  Uint32 this_tick, diff;
  
  this_tick = SDL_GetTicks();

  if(last_tick == 0)
    last_tick = this_tick;

  diff = this_tick - last_tick;
  last_tick = this_tick;

  if ( this_tick < next_tick )
    SDL_Delay(next_tick-this_tick);

  next_tick = this_tick + (1000/FRAMES_PER_SEC);

  return diff;
}

void run_game(char *mapfile)
{
  Uint32 diff=0;
  sim local;
  input_queue local_input;
  keyboard_data local_keyboard;
  int start_time, frame_count=0;
  int total_time=0, output_time=0, ai_time=0, sim_time=0;
  int temp;
  int i, j;
  input_type temp_input;
  map default_map;

  printf("size of sim is %d\n", sizeof(sim));

  keyboard_init(global_option.output);

  clear_sim(&local);

  if(global_option.output)
    output_init(global_option.fullscreen);

  if(global_option.client) {
    if(!client_init(&local, global_option.host, global_option.local_player_count+global_option.local_ai_count)) {
      fprintf(stderr, "Couldn't start client, quitting.\n");
      exit(1);
    }
    okay_to = 0;
  }
  else {

    if(global_option.teams) {
      local.player[1].team_leader = 0;
      local.player[3].team_leader = 2;
    }

    /* insert local players */
    for(i=0; i<global_option.local_player_count; i++)
      have_local_player[i] = 1;

    for(j=0; j<global_option.local_ai_count; j++)
      have_local_player[i+j] = 1;
    
    if(global_option.server) {
      if(!server_init()) {
	fprintf(stderr, "Couldn't start server, quitting.\n");
	exit(1);
      }
    }

    if(!load_map(&default_map, mapfile)) {
      fprintf(stderr, "Map load was unsuccessful. Quitting.\n");
      exit(1);
    }
    
    sim_use_map(&local, &default_map);
  }


  input_queue_clear(&local_input);

  keyboard_data_clear(&local_keyboard);

  /* at this point the sim has all game data */
  
  if(global_option.server) {
    local.export_net_events = 1;
  }

  if(global_option.client)
    local.authority = 0;
  else
    local.authority = 1;

  add_local_players();

  if(global_option.output) {
    output_new_map(&local);
    local.export_output_events = 1;

    keyboard_new_map(&local, &local_keyboard, global_option.output);
  }

  sim_new_game(&local);

  start_time = SDL_GetTicks();
  
  while(1) {

    if(!keyboard_step(&local, &local_input, &local_keyboard, global_option.output, diff))
      break;
    
    frame_count++;

    while(local_input.head != local_input.tail) {
      temp_input = input_dequeue(&local_input);

      if(global_option.client)
	client_send(C_PROCESS_INPUT, temp_input);
      else
	process_input(&local, temp_input);
    }

    if(global_option.output) {
      temp = SDL_GetTicks();
      output_refresh(&local, &local_keyboard);
      output_time += SDL_GetTicks() - temp;
    }

    temp = SDL_GetTicks();
    
    diff = wait_frame();

    if(diff > MAX_STEP_SIZE)
      diff = MAX_STEP_SIZE;

    if(global_option.client)
      if(!client_step(&local, diff))
	break;  /* lost connection */

    /*    if(global_option.client)
      sim_step(&local, max(okay_to-MAX_CLIENT_DELAY-local.elapsed, min(okay_to-local.elapsed, diff)));
      else*/

    if(!global_option.client)
      sim_step(&local, diff);

    if(global_option.server)
      server_step(&local, diff);

    sim_time += SDL_GetTicks() - temp;

    if(global_option.server)
      server_broadcast_okay_to(local.elapsed);

    temp = SDL_GetTicks();

    for(i=0; i<MAX_PLAYER; i++)
      if(have_local_player[i])
	if(local_player_type[i] == AI_PLAYER)
	  ai_go(&local, &local_input, i, 1);

    ai_time += SDL_GetTicks() - temp;

  }

  total_time = SDL_GetTicks() - start_time;

  printf("took %.2f sec, %.1f%% output, %.1f%% ai, %.1f%% sim.\n", ((double)total_time)/1000, ((double)output_time*100)/total_time, ((double)ai_time*100)/total_time, ((double)sim_time*100)/total_time);
  printf(" %.1f fps\n", ((float)frame_count*1000)/total_time);
}

int main(int argc, char *argv[])
{
  int i;
  Uint32 sdl_options;
  char *map = "maps/default.mus";

  global_option.output = 1;
  global_option.fullscreen = 1;

  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i], "--help")) {
      printf("Usage: %s [-s|-c host] [-no] [-nofs] [-hN] [-aiN] [-t] [MAPFILE]\n", argv[0]);
      exit(1);
    }

    if(i==argc-1 && strncmp(argv[i], "-", 1))
      map = argv[argc-1];

    if(!strcmp(argv[i], "--server") || !strcmp(argv[i], "-s")) {
      printf("Running server.\n");
      global_option.server = 1;
    }

    if(i+1<argc)
      if(!strcmp(argv[i], "--client") || !strcmp(argv[i], "-c")) {
	global_option.client = 1;
	global_option.host = argv[i+1];
	printf("Being client to %s.\n", global_option.host);
	i++;
      }

    if(!strcmp(argv[i], "--nooutput") || !strcmp(argv[i], "-no")) {
      printf("Not displaying output.\n");
      global_option.output = 0;
    }

    if(!strcmp(argv[i], "--nofullscreen") || !strcmp(argv[i], "-nofs")) {
      printf("Fullscreen disabled.\n");
      global_option.fullscreen = 0;
    }

    if(!strcmp(argv[i], "--teams") || !strcmp(argv[i], "-t")) {
      printf("Using teams.\n");
      global_option.teams = 1;
    }

    if(!strncmp(argv[i], "-h", 2))
      if(*(argv[i]+2)-'0' <= MAX_PLAYER && *(argv[i]+2)-'0' > 0) {
	printf("Using %d humans.\n", *(argv[i]+2)-'0');
	global_option.local_player_count = *(argv[i]+2)-'0';
      }

    if(!strncmp(argv[i], "-ai", 3))
      if(*(argv[i]+3)-'0' <= MAX_PLAYER && *(argv[i]+3)-'0' > 0) {
	printf("Using %d computers.\n", *(argv[i]+3)-'0');
	global_option.local_ai_count = *(argv[i]+3)-'0';
      }
  }


  sdl_options = SDL_INIT_TIMER;
  sdl_options |= SDL_INIT_VIDEO; /* this is needed to pick up quit msgs from ctrl+c */

  if(global_option.output) {
    sdl_options |= SDL_INIT_AUDIO;
  }

  if(SDL_Init(sdl_options) < 0) {
    fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);

  if(global_option.output) {
    SDL_WM_SetCaption(APPNAME, APPNAME);
  }

  srand(time(NULL));
  
  run_game(map);

  if(global_option.output)
    output_exit();
  
  return 0;
}



